# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Parse electricity generation results from an energy simulation SQL result file.

-
    Args:
        _sql: The file path of the SQL result file that has been generated from
            an energy simulation.

    Returns:
        site_totals: Two numbers indicating the following energy values in kWh.
            _
            * Total on-site produced electricity (postive number)
            * Total on-site electricity consumption (negative number)
        utility_totals: Three numbers indicating the following energy values in kWh.
            _
            * Total on-site produced electricity consumed on-site (positive number)
            * Total on-site produced surplus electricity sold to the utility (positive number)
            * Total electricity purchased from the utility (negative number)
        production: A data collection of all on-site produced electricity (kWh). This
            represents the alternating current (AC) electricity coming out of
            the inverter that processes all on-site power production.
        consumption: A data collection of all on-site consumed electricity (kWh). This
            represents the electrcicity consumed by all heating, cooling, lighting
            equipment, fans, pumps, process loads, and water heaters. All of
            this consumed electricity is assumed to be alternating current (AC).
        dc_power: A list of data collections for the direct current (DC) electricity
            produced by each on-site electricity generator (kWh). Each
            photovoltaic object will have a separate data collection.
"""

ghenv.Component.Name = 'HB Read Generation Result'
ghenv.Component.NickName = 'GenerationResult'
ghenv.Component.Message = '1.7.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '6 :: Result'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

import os
import subprocess
import json
from collections import OrderedDict

try:
    from ladybug.datacollection import HourlyContinuousCollection, \
        MonthlyCollection, DailyCollection
    from ladybug.sql import SQLiteResult
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.result.generation import generation_summary_from_sql, \
        generation_data_from_sql
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# The SQLite3 module doesn't work in IronPython on Mac, so we must make a call
# to the Honeybee CLI (which runs on CPython) to get the results.
def get_results_mac(sql_files):
    from collections import OrderedDict
    cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
            'generation-summary']
    cmds.extend(sql_files)
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, env=custom_env)
    stdout = process.communicate()
    results = json.loads(stdout[0])
    return results


def serialize_data(data_dicts):
    """Reserialize a list of collection dictionaries."""
    if len(data_dicts) == 0 or data_dicts[0] is None:
        return [None] * len(data_dicts)
    elif data_dicts[0]['type'] == 'HourlyContinuous':
        return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'Monthly':
        return [MonthlyCollection.from_dict(data) for data in data_dicts]
    elif data_dicts[0]['type'] == 'Daily':
        return [DailyCollection.from_dict(data) for data in data_dicts]

DC_OUTPUT = 'Generator Produced DC Electricity Energy'
custom_env = os.environ.copy()
custom_env['PYTHONHOME'] = ''


if all_required_inputs(ghenv.Component):
    dc_power = []
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        result_dict = generation_summary_from_sql(_sql)
        production, consumption = generation_data_from_sql(_sql)
        for sql_f in _sql:
            sql_obj = SQLiteResult(sql_f)
            dc_data = sql_obj.data_collections_by_output_name(DC_OUTPUT)
            dc_power.extend(dc_data)

    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        result_dict = get_results_mac(_sql)
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'generation-data']
        cmds.extend(_sql)
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, env=custom_env)
        stdout = process.communicate()
        data_dicts = json.loads(stdout[0])
        production, consumption = serialize_data(data_dicts)

        for sql_f in _sql:
            cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                    'data-by-outputs', sql_f, DC_OUTPUT]
            process = subprocess.Popen(cmds, stdout=subprocess.PIPE, env=custom_env)
            stdout = process.communicate()
            data_dicts = json.loads(stdout[0])
            dc_data = serialize_data(data_dicts[0])
            dc_power.extend(dc_data)

    # output the separate summary results
    site_totals = (
        result_dict['total_production'],
        result_dict['total_consumption']
    )
    utility_totals = (
        result_dict['production_used_on_site'],
        result_dict['production_surplus_sold'],
        result_dict['consumption_purchased']
    )

    # group the generator results by identifier
    dc_dict = OrderedDict()
    for g_data in dc_power:
        gen_id = g_data.header.metadata['System'].split('..')[0]
        g_data.header.metadata['System'] = gen_id
        try:
            dc_dict[gen_id] += g_data
        except KeyError:
            dc_dict[gen_id] = g_data
    dc_power = [dcp for dcp in dc_dict.values()]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Assign photovoltaic properties to a honeybee Shade such that it will generate
electricity in energy simulations.
-

    Args:
        _shades: Honeybee Shades to which photovoltaic properties will be assigned.
        _rated_efficiency_: A number between 0 and 1 for the rated nameplate efficiency
            of the photovoltaic solar cells under standard test conditions (STC).
            _
            Standard test conditions are 1,000 Watts per square meter solar
            irradiance, 25 degrees C cell temperature, and ASTM G173-03 standard
            spectrum. Nameplate efficiencies reported by manufacturers are typically
            under STC.
            _
            Standard poly- or mono-crystalline silicon modules tend to have
            rated efficiencies in the range of 14-17%. Premium high efficiency
            mono-crystalline silicon modules with anti-reflective coatings can
            have efficiencies in the range of 18-20%. Thin film photovoltaic
            modules typically have efficiencies of 11% or less. (Default: 0.15
            for standard silicon solar cells).
        _active_fraction_: The fraction of the parent Shade geometry that is covered in active
            solar cells. This fraction includes the difference between the PV
            panel (aka. PV module) area and the active cells within the panel
            as well as any losses for how the (typically rectangular) panels
            can be arranged on the Shade geometry. When the parent Shade
            geometry represents just the solar panels, this fraction is typically
            around 0.9 given that the framing elements of the panel reduce the
            overall active area. (Default: 0.9).
        _module_type_: Text or an integer to indicate the type of solar module. This is used
            to determine the temperature coefficients used in the simulation of the
            photovoltaic modules. Choose from the three options below. If unspecified,
            the module_type will be inferred from the rated_efficiency of these
            PVProperties using the rated efficiencies listed below.
            _
                * 0 - Standard - 12% <= rated_efficiency < 18%
                * 1 - Premium - rated_efficiency >= 18%
                * 2 - ThinFilm - rated_efficiency < 12%
        _mounting_type_: Text or an integer to indicate the type of mounting and/or tracking used
            for the photovoltaic array. Note that the OneAxis options have an axis
            of rotation that is determined by the azimuth of the parent Shade
            geometry. Also note that, in the case of one or two axis tracking,
            shadows on the (static) parent Shade geometry still reduce the
            electrical output, enabling the simulation to account for large
            context geometry casting shadows on the array. However, the effects
            of smaller detailed shading may be improperly accounted for and self
            shading of the dynamic panel geometry is only accounted for via the
            tracking_gcr property. Choose from the following. (Default: FixedOpenRack).
            _
                * 0 - FixedOpenRack - ground or roof mounting where the air flows freely 
                * 1 - FixedRoofMounted - mounting flush with the roof with limited air flow
                * 2 - OneAxis - a fixed tilt and azimuth, which define an axis of rotation
                * 3 - OneAxisBacktracking - same as OneAxis but with controls to reduce self-shade
                * 4 - TwoAxis - a dynamic tilt and azimuth that track the sun
        _loss_fraction_: A number between 0 and 1 for the fraction of the electricity output
            lost due to factors other than EPW climate conditions, panel
            efficiency/type, active area, mounting, and inverter conversion
            from DC to AC.
            _
            Factors that should be accounted for in this input include soiling,
            snow, wiring losses, electrical connection losses, manufacturer
            defects/tolerances/mismatch in cell characteristics, losses from power
            grid availability, and losses due to age or light-induced degradation.
            _
            Losses from these factors tend to be between 10-20% but can vary widely
            depending on the installation, maintenance and the grid to which the
            panels are connected. The loss_fraction_from_components staticmethod
            on this class can be used to estimate this value from the various
            factors that it is intended to account for. (Default: 0.14).
        tracking_gcr_: A number between 0 and 1 that ONLY APPLIES TO ARRAYS WITH ONE AXIS
            mounting_type.
            _
            The ground coverage ratio (GCR) is the ratio of module surface area
            to the area of the ground beneath the array, which is used to account
            for self shading of single-axis panels as they move to track the sun.
            A GCR of 0.5 means that, when the modules are horizontal, half of the
            surface below the array is occupied by the array. An array with wider
            spacing between rows of modules has a lower GCR than one with narrower
            spacing. A GCR of 1 would be for an array with no space between modules,
            and a GCR of 0 for infinite spacing between rows. Typical values
            range from 0.3 to 0.6. (Default: 0.4).
        _name_: An optional text name for the photovoltaic properties. This can be useful
            for keeping track of different photovoltaics when using several of
            these components. If unspecified, a unique one will be generated.

    Returns:
        report: Reports, errors, warnings, etc.
        shades: The input Shades with photovoltaic properties assigned.
"""

ghenv.Component.Name = 'HB Photovoltaic Properties'
ghenv.Component.NickName = 'Photovoltaic'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

import math

try:  # import the ladybug geometry dependencies
    from ladybug_geometry.geometry3d import Vector3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import honeybee_energy dependencies
    from honeybee_energy.generator.pv import PVProperties
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import angle_tolerance, conversion_to_meters
    from ladybug_rhino.grasshopper import all_required_inputs, \
        document_counter, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# dictionaries to convert user input to enumerations
MODULE_TYPES = {
    'standard': 'Standard',
    'premium': 'Premium',
    'thinfilm': 'ThinFilm',
    '0': 'Standard',
    '1': 'Premium',
    '2': 'Standard'
}
MOUNTING_TYPES = {
    'fixedopenrack': 'FixedOpenRack',
    'fixedroofmounted': 'FixedRoofMounted',
    'oneaxis': 'OneAxis',
    'oneaxisbacktracking': 'OneAxisBacktracking',
    'twoaxis': 'TwoAxis',
    '0': 'FixedOpenRack',
    '1': 'FixedRoofMounted',
    '2': 'OneAxis',
    '3': 'OneAxisBacktracking',
    '4': 'TwoAxis',
}


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    shades = []
    for shade in _shades:
        shades.append(shade.duplicate())
        if math.degrees(Vector3D(0, 0, 1).angle(shade.normal)) - 90 > angle_tolerance:
            msg = 'Shade "{}" is pointing downwards, which is atypical of photovoltaics.\n' \
                'You will likely want to flip the geometry to have it point upwards to ' \
                'the sky.'.format(shade.display_name)
            print(msg)
            give_warning(ghenv.Component, msg)

    # assign defaults for the PV properties
    display_name = 'Photovoltaic Array {}'.format(document_counter('pv_count')) \
        if _name_ is None else _name_
    pv_id = clean_ep_string(display_name)
    eff = _rated_efficiency_ if _rated_efficiency_ is not None else 0.15
    act_fraction = _active_fraction_ if _active_fraction_ is not None else 0.9
    mod_type = MODULE_TYPES[_module_type_.lower()] \
        if _module_type_ is not None else None
    mount_type = MOUNTING_TYPES[_mounting_type_.lower()] \
        if _mounting_type_ is not None else 'FixedOpenRack'
    loss_fraction = _loss_fraction_ if _loss_fraction_ is not None else 0.14
    gcr = tracking_gcr_ if tracking_gcr_ is not None else 0.4

    # create the base PV properties
    pv_props = PVProperties(pv_id, eff, act_fraction, mod_type, mount_type,
                            loss_fraction, gcr)
    pv_props.display_name = pv_id

    # assign the PV properties to the Shades
    conversion = conversion_to_meters() ** 2
    total_area = 0
    for shade in shades:
        shade.properties.energy.pv_properties = pv_props
        total_area += shade.area * conversion

    # compute the rated DC size from the area and other inputs
    dc_size = int(total_area * eff * act_fraction * 1000)
    msg = 'The combined rated DC size across all input shades is {:,} W.'.format(dc_size)
    print(msg)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize the desirability of shade in terms of its impact on the annual heating
and cooling loads of Honeybee Rooms.
_
The calculation runs by performing a simple fast energy simulation of the
connected Honeybee Rooms without any shade other than context. The resulting
heating/cooling loads are extracted at each timestep of the simulation along with
the direct (beam) solar gain through each of the Room's windows. Solar vectors
are generated for each step of the simulation and projected from the Room's
Aperture geometries through the shades assigned to those Apertures.
_
Solar vectors for timesteps when the Room is cooling mode contribute positively
to shade desirability (shade help) while solar vectors for hours when the
Room is heating mode contribute negatively (shade harm). This contribution
is weighted by how much cooling or heating energy the Room requires at the
timestep along with the direct solar gain through each Aperture at the timestep.
_
The component outputs a colored mesh of the shades assigned to the Room Apertures
illustrating the net effect of shading each part of the geometry. A higher
saturation of blue indicates that shading the cell is desirable. A higher
saturation of red indicates that shading the cell is harmful (blocking more sun
when the Room is in heating mode than cooling mode). Desaturated cells indicate
that shading the cell will have relatively little effect on the heating or
cooling loads of the Room.
_
The units for shade desirability are kWh of Room cooling load avoided per unit
area of shade if the test cell of the shade is helpful (blue). If the test cell
is harmful (red), the units are kWh of Room heating load increased per unit
area of shade. So, if a given square meter of input _shade_geo has a shad
desirability of 10 kWh/m2, this means that a shade in this location provides
roughly 10 kWh of avoided cooling load to the parent Room over the year.
_
The method used by this component is based off of the Shaderade method developed
by Jon Sargent, Jeffrey Niemasz and Christoph Reinhart. More information can be
found in the following publication document:
Sargent, Jon; Niemasz, Jeffrey; Reinhart, Christoph. SHADERADE: Combining
Rhinoceros and EnergyPlus for the Design of Static Exterior Shading Devices.
Building Simulation, 2011, Sydney, Australia.
http://www.ibpsa.org/proceedings/bs2011/p_1209.pdf
-

    Args:
        _rooms: A list of Honeybee Rooms for which cooling/heating shade benefit/harm will
            be evaluated. At least some of these Rooms should have Apertures
            with Shades assigned to them in order for this component to produce
            meaningful results. Note that all Shades generated with the "HB Louver
            Shades" component or the "HB Extruded Border" component are
            automatically assigned to a parent Aperture. For more complex
            Shade geometries, the "HB Add Shade" component can be used to
            assign the Shade to a parent Aperture.
        context_: Honeybee Shades representing context geometry that can block sun
            to the _rooms, therefore discounting any benefit or harm that could
            come to the Room's Shades.
        _epw_file: Path to an .epw file on your system as a text string. This will be
            used in the energy simulation to determine heating/cooling loads
            and to generate solar vectors for the shade benefit calculation.
        _north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. (Default: 0).
        _grid_size: A positive number in Rhino model units for the size of grid cells at
            which the Shade geometries of the input _rooms will be subdivided
            for shade benefit analysis. The smaller the grid size, the higher
            the resolution of the analysis and the longer the calculation will
            take. So it is recommended that one start with a large value here
            and decrease the value as needed. However, the grid size should
            usually be smaller than the dimensions of the smallest piece of
            Shade geometry in order to yield meaningful results.
        _timestep_: An integer for the number of timesteps per hour at which the energy
            simulation will run and sun vectors will be generated for the analysis.
            Higher values will result in the generation of more vectors, which
            will make the resulting shade meshes smoother and produce a better
            representation of real benefit/harm. However, the calculation will take
            longer as there are more intersection operations to perform. The
            default is 1 timestep per hour, which is the coarsest resolution
            avalable and the fastest calculation.
        lag_time_: A number for the amount of time in hours between when solar gain
            eneters the room and the gain results in an increased cooling load.
            Typically, it takes an hour or so for solar gains falling on the
            room floors to heat up the floor surface and then convect to the
            room air where the gain can be absorbed by a cooling system. This
            means that the cooling value associated with each sun vector should
            be a step or two after the time of the sun vector. Lag time can
            be longer than an hour if the room has a particularly high thermal
            mass or it may be shorter if the room has less mass or uses a radiant
            cooling system integrated into the floor where the sun is absorbed.
            Note that the value input here can be a decimal value to indicate
            that the lag time is a fraction of an hour. (Default: 1.0 hour).
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        _cpu_count_: An integer to set the number of CPUs used in the execution of the
            intersection calculation. If unspecified, it will automatically default
            to one less than the number of CPUs currently available on the
            machine or 1 if only one processor is available.
        _run: Set to "True" to run the component and perform shade benefit analysis.

    Returns:
        report: ...
        vectors: The sun vectors that were used to evaluate the shade (note that
            these will increase as the _timestep_ increases).
        points: Points across the room Aperture geometry from which sun vectors
            are projected. Note that only Apertures with assigned Shades are
            evaluated in order to avoid unnessarily increasing the calculation
            time by evaluating windows for which there is not shade.
        mesh: A colored mesh of the Shades assigned to the room's apertures showing
            where shading is helpful (blue), harmful (red), or does not make
            much of a difference (white or desaturated colors). Note that
            the colors can change depending upon the input legend_par_.
        legend: Legend showing the numeric values of kWh per unit shade area of
            decreased/increased cooling/heating load that correspond to the
            colors in the shade mesh.
        title: A text object for the study title.
        shade_help: The cumulative kWh of avoided cooling load per square area unit
            obtained by shading the given cell. If a given square meter of
            shade geometry has a helpfulness of 10 kWh/m2, this means that
            a shade in this location decreases the cooling load of the Room
            by roughly 10 kWh over the year.
        shade_harm: The cumulative kWh of increased heating load per square area unit
            obtained by shading the given cell. If a given square meter of
            shade geometry has a harmfulness of -10 kWh/m2, this means that
            a shade in this location increases the heating load of the Room
            by roughly 10 kWh over the year.
        shade_net: The sum of the helpfulness and harmfulness for each cell. This will be
            negative if shading the cell has a net harmful effect and positive
            if the shade has a net helpful effect.
"""

ghenv.Component.Name = 'HB Load Shade Benefit'
ghenv.Component.NickName = 'LoadShadeBenefit'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os
import subprocess
import json
import math

try:
    from ladybug_geometry.geometry3d import Mesh3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.futil import write_to_file_by_name, nukedir
    from ladybug.sunpath import Sunpath
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
    from ladybug.epw import EPW
    from ladybug.sql import SQLiteResult
    from ladybug.datacollection import HourlyContinuousCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
    from honeybee.boundarycondition import Outdoors
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.simulation.parameter import SimulationParameter
    from honeybee_energy.run import run_idf
    from honeybee_energy.result.err import Err
    from honeybee_energy.writer import energyplus_idf_version
    from honeybee_energy.config import folders as energy_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from lbt_recipes.version import check_energyplus_version
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector2d, to_joined_gridded_mesh3d
    from ladybug_rhino.fromgeometry import from_face3d, from_mesh3d, \
        from_point3d, from_vector3d
    from ladybug_rhino.config import conversion_to_meters, units_system, \
        tolerance, angle_tolerance, units_abbreviation
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, generate_intersection_rays, \
        intersect_rays_with_mesh_faces
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning, hide_output, \
        recommended_processor_count
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def serialize_data(data_dicts):
    """Reserialize a list of HourlyContinuousCollection dictionaries."""
    return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]

# List of all the output strings that will be requested
cool_out = 'Zone Ideal Loads Supply Air Total Cooling Energy'
heat_out = 'Zone Ideal Loads Supply Air Total Heating Energy'
solar_out = 'Surface Window Transmitted Beam Solar Radiation Energy'
all_output = (cool_out, heat_out, solar_out)


if all_required_inputs(ghenv.Component) and _run:
    # check the presence of energyplus and check that the version is compatible
    check_energyplus_version()

    # set the defaults and process all of the inputs
    workers = _cpu_count_ if _cpu_count_ is not None else recommended_processor_count()
    timestep = _timestep_ if _timestep_ is not None else 1
    lag_time = 1 if lag_time_ is None else lag_time_
    lag_steps = int(timestep * lag_time)
    if _north_ is not None:  # process the north_
        try:
            _north_ = math.degrees(
                to_vector2d(_north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            _north_ = float(north_)
    else:
        _north_ = 0

    # gather all assigned shades and remove them from the rooms
    rooms = [r.duplicate() for r in _rooms]  # duplicate to avoid editing input
    ap_count, shd_count = 0, 0
    shade_dict = {}
    for room in rooms:
        if room.properties.energy.is_conditioned and \
                room.properties.energy.setpoint is not None:
            r_dict = {}
            for face in room.faces:
                if isinstance(face.boundary_condition, Outdoors):
                    aps = face.apertures
                    if len(aps) != 0:
                        fap_ids, fap_geos, fshd_geos = [], [], []
                        for ap in aps:
                            fap_ids.append(ap.identifier.upper())
                            fap_geos.append(from_face3d(ap.geometry))
                            fshd_geos.extend(from_face3d(shd.geometry)
                                             for shd in ap.outdoor_shades)
                            ap.remove_shades()  # remove shades for the energy simulation
                        if len(fshd_geos) != 0:
                            r_dict[face.identifier] = {
                                'ap_ids': fap_ids,
                                'ap_geo': fap_geos,
                                'shd_geo': fshd_geos,
                                'normal': from_vector3d(face.normal)
                            }
                            ap_count += len(fap_geos)
                            shd_count += len(fshd_geos)
            if len(r_dict) != 0:
                shade_dict[room.identifier.upper()] = r_dict

    # make sure that there are shades to evaluate
    if shd_count == 0:
        msg = 'There were no shades to evaluate across all of the input _rooms.\n' \
            'Make sure that shades are assigned to the apertures of conditioned\n' \
            'rooms or use components like "HB Louver Shades" to generate shades\n' \
            'that are assigned to the apertures.'
        print(msg)
        raise ValueError(msg)
    else:
        msg = 'Evaluating {} shade geometries across {} apertures.'.format(
            shd_count, ap_count)
        print(msg)

    # create the Model from the _rooms and context_
    _model = Model('Load_Shade_Benefit', rooms, orphaned_shades=context_,
                   units=units_system(),
                   tolerance=tolerance, angle_tolerance=angle_tolerance)

    # process the simulation folder name and the directory
    directory = os.path.join(folders.default_simulation_folder, _model.identifier)
    sch_directory = os.path.join(directory, 'schedules')
    nukedir(directory)  # delete any existing files in the directory

    # create simulation parameters for the coarsest/fastest E+ sim possible
    _sim_par_ = SimulationParameter()
    _sim_par_.timestep = timestep
    _sim_par_.north_angle = _north_
    _sim_par_.shadow_calculation.solar_distribution = 'FullExterior'
    _sim_par_.output.reporting_frequency = 'Timestep'
    _sim_par_.output.include_html = False
    for out_p in all_output:
        _sim_par_.output.add_output(out_p)

    # assign design days from the EPW
    msg = None
    folder, epw_file_name = os.path.split(_epw_file)
    ddy_file = os.path.join(folder, epw_file_name.replace('.epw', '.ddy'))
    if os.path.isfile(ddy_file):
        try:
            _sim_par_.sizing_parameter.add_from_ddy_996_004(ddy_file)
        except AssertionError:
            msg = 'No design days were found in the .ddy file next to the _epw_file.'
    else:
         msg = 'No .ddy file was found next to the _epw_file.'
    if msg is not None:
        epw_obj = EPW(_epw_file)
        des_days = [epw_obj.approximate_design_day('WinterDesignDay'),
                    epw_obj.approximate_design_day('SummerDesignDay')]
        _sim_par_.sizing_parameter.design_days = des_days
        msg = msg + '\nDesign days were generated from the input _epw_file but this ' \
            '\nis not as accurate as design days from DDYs distributed with the EPW.'
        give_warning(ghenv.Component, msg)
        print(msg)

    # create the strings for simulation paramters and model
    ver_str = energyplus_idf_version() if energy_folders.energyplus_version \
        is not None else energyplus_idf_version(compatibe_ep_version)
    sim_par_str = _sim_par_.to_idf()
    model_str = _model.to.idf(
        _model, schedule_directory=sch_directory, patch_missing_adjacencies=True)
    idf_str = '\n\n'.join([ver_str, sim_par_str, model_str])

    # write the final string into an IDF
    idf = os.path.join(directory, 'in.idf')
    write_to_file_by_name(directory, 'in.idf', idf_str, True)

    # run the IDF through EnergyPlus
    silent = True if _run == 1 else False
    sql, zsz, rdd, html, err = run_idf(idf, _epw_file, silent=silent)
    if sql is None and err is not None:  # something went wrong; parse the errors
        err_obj = Err(err)
        print(err_obj.file_contents)
        for error in err_obj.fatal_errors:
            raise Exception(error)

    # parse the result sql and get the timestep data collections
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(sql)
        cooling = sql_obj.data_collections_by_output_name(cool_out)
        heating = sql_obj.data_collections_by_output_name(heat_out)
        solar = sql_obj.data_collections_by_output_name(solar_out)
    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'data-by-outputs', sql]
        for outp in all_output:
            cmds.append('["{}"]'.format(outp))
        custom_env = os.environ.copy()
        custom_env['PYTHONHOME'] = ''
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, env=custom_env)
        stdout = process.communicate()
        data_coll_dicts = json.loads(stdout[0])
        cooling = serialize_data(data_coll_dicts[0])
        heating = serialize_data(data_coll_dicts[1])
        solar = serialize_data(data_coll_dicts[2])

    # convert the results to a dictionary for quick access
    cool_dict, heat_dict, solar_dict = {}, {}, {}
    for cool in cooling:
        cool_dict[cool.header.metadata['System'].split(' ')[0]] = cool
    for heat in heating:
        heat_dict[heat.header.metadata['System'].split(' ')[0]] = heat
    for sol in solar:
        solar_dict[sol.header.metadata['Surface']] = sol

    # initialize sunpath based on the EPW and get all of the vectors
    epw_obj = EPW(_epw_file)
    location = epw_obj.location
    sp = Sunpath.from_location(location, _north_)
    lb_vecs, relevant_i = [], []
    for i, dt in enumerate(solar[0].datetimes):
        sun = sp.calculate_sun_from_date_time(dt)
        if sun.is_during_day:
            lb_vecs.append(sun.sun_vector_reversed)
            relevant_i.append(i)
    vectors = [from_vector3d(lb_vec) for lb_vec in lb_vecs]

    # if there is context, remove any rays that are blocked by the context
    context_mesh = None
    if len(context_) != 0 and context_[0] is not None:
        context_mesh = join_geometry_to_mesh([from_face3d(c.geometry) for c in context_])

    # loop through the relevant rooms and compute shade benefit
    points, mesh = [], []
    shade_help, shade_harm, shade_net = [], [], []
    hide_output(ghenv.Component, 2)
    for room_id, room_data in shade_dict.items():
        cool_vals = cool_dict[room_id].values
        heat_vals = heat_dict[room_id].values
        # shif the values by the lag
        cool_vals = cool_vals[-lag_steps:] + cool_vals[:-lag_steps]
        heat_vals = heat_vals[-lag_steps:] + heat_vals[:-lag_steps] 
        for ap_data in room_data.values():
            solar_vals = solar_dict[ap_data['ap_ids'][0]]
            for ap_id in ap_data['ap_ids'][1:]:
                solar_vals += solar_dict[ap_id]

            # create the gridded mesh from the geometry
            analysis_mesh = to_joined_gridded_mesh3d(ap_data['shd_geo'], _grid_size)
            ap_mesh = from_mesh3d(analysis_mesh)
            study_mesh = to_joined_gridded_mesh3d(ap_data['ap_geo'], _grid_size / 1.75)
            ap_points = [from_point3d(pt) for pt in study_mesh.face_centroids]
            points.extend(ap_points)
            mesh.append(analysis_mesh)

            # create a series of rays that represent the sun projected through the shade
            int_rays = generate_intersection_rays(ap_points, vectors)
            normals = [ap_data['normal']] * len(int_rays)

            # intersect the sun rays with the shade mesh
            face_int = intersect_rays_with_mesh_faces(
                ap_mesh, int_rays, context_mesh, normals, cpu_count=workers)

            # loop through the face intersection result and evaluate the benefit
            pt_div = 1 / float(len(ap_points))
            for face_res, face_area in zip(face_int, analysis_mesh.face_areas):
                f_help, f_harm = 0, 0
                for t_ind in face_res:
                    ri = relevant_i[t_ind]
                    cl, ht, sl = cool_vals[ri], heat_vals[ri], solar_vals[ri]
                    if cl > 0:  # a step where shade helps
                        f_help += min(cl, sl)
                    elif ht > 0:
                        f_harm -= min(ht, sl)
                # Normalize by the area of the cell so there's is a consistent metric
                # between cells of different areas.
                shd_help = ((f_help / face_area)) * pt_div
                shd_harm = ((f_harm / face_area)) * pt_div
                shade_help.append(shd_help)
                shade_harm.append(shd_harm)
                shade_net.append(shd_help + shd_harm)

    # create the mesh and legend outputs
    mesh = Mesh3D.join_meshes(mesh)
    graphic = GraphicContainer(shade_net, mesh.min, mesh.max, legend_par_)
    graphic.legend_parameters.title = 'kWh/{}2'.format(units_abbreviation())
    if legend_par_ is None or legend_par_.are_colors_default:
        graphic.legend_parameters.colors = reversed(Colorset.shade_benefit_harm())
    if legend_par_ is None or legend_par_.min is None or legend_par_.max is None:
        bnd_val = max(max(shade_net), abs(min(shade_net)))
        if legend_par_ is None or legend_par_.min is None:
            graphic.legend_parameters.min = -bnd_val
        if legend_par_ is None or legend_par_.max is None:
            graphic.legend_parameters.max = bnd_val
    title = text_objects('Cooling/Heating Load Shade Benefit', graphic.lower_title_location,
                         graphic.legend_parameters.text_height * 1.5,
                         graphic.legend_parameters.font)

    # create all of the visual outputs
    mesh.colors = graphic.value_colors
    mesh = from_mesh3d(mesh)
    legend = legend_objects(graphic.legend)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Apply values for setpoints to a Room or ProgramType.
-

    Args:
        _room_or_program: Honeybee Rooms or ProgramType objects to which the input
            setpoints should be assigned. This can also be the identifier of a
            ProgramType to be looked up in the program type library. This can
            also be a Honeybee Model for which all Rooms will be assigned
            the setpoints.
        cooling_setpt_: A numerical value for a single constant temperature for
            the cooling setpoint [C].
        heating_setpt_: A numerical value for a single constant temperature for
            the heating setpoint [C].
        humid_setpt_: A numerical value for a single constant value for the
            humidifying setpoint [%].
        dehumid_setpt_: A numerical value for a single constant value for the
            dehumidifying setpoint [%].
        cutout_difference_: An optional positive number for the temperature difference between the
            cutout temperature and the setpoint temperature. Specifying a non-zero
            number here is useful for modeling the throttling range associated
            with a given setup of setpoint controls and HVAC equipment. Throttling
            ranges describe the range where a zone is slightly over-cooled or
            over-heated beyond the thermostat setpoint. They are used to avoid
            situations where HVAC systems turn on only to turn off a few minutes later,
            thereby wearing out the parts of mechanical systems faster. They can
            have a minor impact on energy consumption and can often have significant
            impacts on occupant thermal comfort, though using the default value
            of zero will often yield results that are close enough when trying
            to estimate the annual heating/cooling energy use. Specifying a value
            of zero effectively assumes that the system will turn on whenever
            conditions are outside the setpoint range and will cut out as soon
            as the setpoint is reached. (Default: 0).

    Returns:
        report: Reports, errors, warnings, etc.
        mod_obj: The input Rooms or ProgramTypes with their setpoint values edited.
"""

ghenv.Component.Name = "HB Apply Setpoint Values"
ghenv.Component.NickName = 'ApplySetpointVals'
ghenv.Component.Message = '1.7.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = "2"

import uuid

try:
    from honeybee.model import Model
    from honeybee.room import Room
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.setpoint import Setpoint
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
    import honeybee_energy.lib.scheduletypelimits as _type_lib
    from honeybee_energy.lib.programtypes import program_type_by_identifier, \
        building_program_type_by_identifier
    from honeybee_energy.programtype import ProgramType
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))
try:
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def dup_setpoint(hb_obj):
    """Duplicate a setpoint object assigned to a Room or ProgramType."""
    # try to get the setpoint object assgined to the Room or ProgramType
    try:  # assume it's a Room
        setpt_obj = hb_obj.properties.energy.setpoint
    except AttributeError:  # it's a ProgramType
        setpt_obj = hb_obj.setpoint

    load_id = '{}_Setpoint'.format(hb_obj.identifier)
    try:  # duplicate the setpoint object
        dup_load = setpt_obj.duplicate()
        dup_load.identifier = load_id
        return dup_load
    except AttributeError:  # create a new object if it does not exist
        heat_sch = ScheduleRuleset.from_constant_value(
            '{}_HtgSetp'.format(hb_obj.identifier), -50, _type_lib.temperature)
        cool_sch = ScheduleRuleset.from_constant_value(
            '{}_ClgSetp'.format(hb_obj.identifier), 50, _type_lib.temperature)
        return Setpoint(load_id, heat_sch, cool_sch)


def assign_setpoint(hb_obj, setpt_obj):
    """Assign a setpoint object to a Room or a ProgramType."""
    try:  # assume it's a Room
        hb_obj.properties.energy.setpoint = setpt_obj
    except AttributeError:  # it's a ProgramType
        hb_obj.setpoint = setpt_obj


def duplicate_and_id_program(program):
    """Duplicate a program and give it a new unique ID."""
    new_prog = program.duplicate()
    new_prog.identifier = '{}_{}'.format(program.identifier, str(uuid.uuid4())[:8])
    return new_prog


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    mod_obj, edit_objs = [], []
    for obj in _room_or_program:
        if isinstance(obj, Room):
            new_obj = obj.duplicate()
            mod_obj.append(new_obj)
            edit_objs.append(new_obj)
        elif isinstance(obj, Model):
            new_obj = obj.duplicate()
            mod_obj.append(new_obj)
            edit_objs.extend(new_obj.rooms)
        elif isinstance(obj, ProgramType):
            new_obj = duplicate_and_id_program(obj)
            mod_obj.append(new_obj)
            edit_objs.append(new_obj)
        elif isinstance(obj, str):
            try:
                program = building_program_type_by_identifier(obj)
            except ValueError:
                program = program_type_by_identifier(obj)
            new_obj = duplicate_and_id_program(program)
            mod_obj.append(new_obj)
            edit_objs.append(new_obj)
        else:
            raise TypeError('Expected Honeybee Room, Model or ProgramType. '
                            'Got {}.'.format(type(obj)))

    # assign the cooling_setpt_
    if len(cooling_setpt_) != 0:
        for i, obj in enumerate(edit_objs):
            setpoint = dup_setpoint(obj)
            setpoint.cooling_setpoint = longest_list(cooling_setpt_, i)
            assign_setpoint(obj, setpoint)

    # assign the heating_setpt_
    if len(heating_setpt_) != 0:
        for i, obj in enumerate(edit_objs):
            setpoint = dup_setpoint(obj)
            setpoint.heating_setpoint = longest_list(heating_setpt_, i)
            assign_setpoint(obj, setpoint)

    # assign the humid_setpt_
    if len(humid_setpt_) != 0:
        for i, obj in enumerate(edit_objs):
            setpoint = dup_setpoint(obj)
            setpoint.humidifying_setpoint = longest_list(humid_setpt_, i)
            assign_setpoint(obj, setpoint)

    # assign the dehumid_setpt_
    if len(dehumid_setpt_) != 0:
        for i, obj in enumerate(edit_objs):
            setpoint = dup_setpoint(obj)
            setpoint.dehumidifying_setpoint = longest_list(dehumid_setpt_, i)
            assign_setpoint(obj, setpoint)

    # assign the cutout_difference_
    if len(cutout_difference_) != 0:
        for i, obj in enumerate(edit_objs):
            setpoint = dup_setpoint(obj)
            setpoint.setpoint_cutout_difference = longest_list(cutout_difference_, i)
            assign_setpoint(obj, setpoint)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Run Honeybee objects capable of generating electricity (such as Shades with PV
properties) through a quick energy simulation to obtain an estimate of
electricity production.
_
Note that this component only evaluates electricity production and not energy
consumption. Any number of Honeybee Rooms or other objects can be connected
but they will only be simulated as context shade that casts shadows on the
generator objects.
-

    Args:
        _hb_objs: An array of honeybee Rooms, Faces, Apertures, Doors or Shades to be
            included in the simulation of electricity production. This can also
            be an entire Model to be simulated. Any number of Honeybee Rooms or
            non-generating objects can be connected but they will only be
            simulated as context shade that casts shadows on the
            generator objects.
        _epw_file: Path to an .epw file on your system as a text string.
        _north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. (Default: 0).
        _inverter_eff_: A number between 0 and 1 for the load centers's inverter nominal
            rated DC-to-AC conversion efficiency. An inverter converts DC power,
            such as that output by photovoltaic panels, to AC power, such as
            that distributed by the electrical grid and is available from
            standard electrical outlets. Inverter efficiency is defined as the
            inverter's rated AC power output divided by its rated DC power
            output. (Default: 0.96).
        _dc_to_ac_size_: A positive number (typically greater than 1) for the ratio of the
            inverter's DC rated size to its AC rated size. Typically, inverters
            are not sized to convert the full DC output under standard test
            conditions (STC) as such conditions rarely occur in reality and
            therefore unnecessarily add to the size/cost of the inverter. For a
            system with a high DC to AC size ratio, during times when the 
            DC power output exceeds the inverter's rated DC input size, the inverter
            limits the array's power output by increasing the DC operating voltage,
            which moves the arrays operating point down its current-voltage (I-V)
            curve. The default value of 1.1 is reasonable for most systems. A
            typical range is 1.1 to 1.25, although some large-scale systems have
            ratios of as high as 1.5. The optimal value depends on the system's
            location, array orientation, and module cost. (Default: 1.1).
        _run: Set to "True" to run the simulation to obtain annual loads. This can
            also be the integer 2 to run the simulation while being able to see
            the simulation process (with a batch window).

    Returns:
        report: A report of the energy simulation run.
        total_ac: A number for the total on-site produced alternating current (AC)
            electricity in kWh.
        ac_power: A data collection of all on-site produced electricity (kWh). This
            represents the alternating current (AC) electricity coming out of
            the inverter that processes all on-site power production.
        generators: A list of names for each of the electricity generation objects that
            were found among the connected _hb_objs. These names align with the
            tota_dc output below as well as the dc_power data collections.
        total_dc: A list of numbers for the direct current (DC) electricity produced
            by each generator object in kWh.
        dc_power: A list of data collections for the direct current (DC) electricity
            produced by each on-site electricity generator (kWh). Each
            photovoltaic object will have a separate data collection.
"""

ghenv.Component.Name = 'HB Generation Loads'
ghenv.Component.NickName = 'GenLoads'
ghenv.Component.Message = '1.7.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '5 :: Simulate'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os
import subprocess
import json
from collections import OrderedDict

try:
    from ladybug.futil import write_to_file_by_name, nukedir
    from ladybug.sql import SQLiteResult
    from ladybug.datacollection import HourlyContinuousCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee.config import folders
    from honeybee.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.simulation.parameter import SimulationParameter
    from honeybee_energy.run import run_idf
    from honeybee_energy.result.err import Err
    from honeybee_energy.writer import energyplus_idf_version
    from honeybee_energy.config import folders as energy_folders
    from honeybee_energy.lib.constructions import opaque_construction_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from lbt_recipes.version import check_energyplus_version
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector2d
    from ladybug_rhino.config import units_system, tolerance, angle_tolerance
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def serialize_data(data_dicts):
    """Reserialize a list of HourlyContinuousCollection dictionaries."""
    return [HourlyContinuousCollection.from_dict(data) for data in data_dicts]


# List of all the output strings that will be requested
ac_out = 'Facility Total Produced Electricity Energy'
dc_out = 'Generator Produced DC Electricity Energy'
energy_output = (ac_out, dc_out)


if all_required_inputs(ghenv.Component) and _run:
    # check the presence of energyplus and check that the version is compatible
    check_energyplus_version()

    # create the Model from the _hb_objs
    models = [obj.duplicate() for obj in _hb_objs if isinstance(obj, Model)]
    other_objs = [obj.duplicate() for obj in _hb_objs if not isinstance(obj, Model)]
    model = Model.from_objects('Generation_Loads', other_objs, units_system(),
                               tolerance, angle_tolerance)
    for m in models:
        model.add_model(m)
    model.rooms_to_orphaned()
    soil_constr = opaque_construction_by_identifier('Mud')
    model.properties.energy.generate_ground_room(soil_constr)

    # set the inverter efficiency and size
    if _inverter_eff_ is not None:
        model.properties.energy.electric_load_center.inverter_efficiency = _inverter_eff_
    if _dc_to_ac_size_ is not None:
        model.properties.energy.electric_load_center.inverter_dc_to_ac_size_ratio = _dc_to_ac_size_

    # process the simulation folder name and the directory
    directory = os.path.join(folders.default_simulation_folder, model.identifier)
    sch_directory = os.path.join(directory, 'schedules')
    nukedir(directory)  # delete any existing files in the directory

    # create simulation parameters for the coarsest/fastest E+ sim possible
    _sim_par_ = SimulationParameter()
    _sim_par_.timestep = 6
    _sim_par_.shadow_calculation.solar_distribution = 'FullExteriorWithReflections'
    _sim_par_.output.reporting_frequency = 'Hourly'
    _sim_par_.output.add_electricity_generation()
    _sim_par_.output.include_html = False
    _sim_par_.simulation_control.do_zone_sizing = False
    _sim_par_.simulation_control.do_system_sizing = False
    _sim_par_.simulation_control.do_plant_sizing = False

    # set the north if it is not defaulted
    if _north_ is not None:
        try:
            _sim_par_.north_vector = to_vector2d(_north_)
        except AttributeError:  # north angle instead of vector
            _sim_par_.north_angle = float(_north_)

    # create the strings for simulation paramters and model
    ver_str = energyplus_idf_version() if energy_folders.energyplus_version \
        is not None else energyplus_idf_version(compatibe_ep_version)
    sim_par_str = _sim_par_.to_idf()
    model_str = model.to.idf(
        model, schedule_directory=sch_directory, patch_missing_adjacencies=True)
    idf_str = '\n\n'.join([ver_str, sim_par_str, model_str])

    # write the final string into an IDF
    idf = os.path.join(directory, 'in.idf')
    write_to_file_by_name(directory, 'in.idf', idf_str, True)

    # run the IDF through EnergyPlus
    silent = True if _run == 1 else False
    sql, zsz, rdd, html, err = run_idf(idf, _epw_file, silent=silent)
    if sql is None and err is not None:  # something went wrong; parse the errors
        err_obj = Err(err)
        print(err_obj.file_contents)
        for error in err_obj.fatal_errors:
            raise Exception(error)

    # parse the result sql and get the monthly data collections
    if os.name == 'nt':  # we are on windows; use IronPython like usual
        sql_obj = SQLiteResult(sql)
        ac_power = sql_obj.data_collections_by_output_name(ac_out)
        dc_power = sql_obj.data_collections_by_output_name(dc_out)
    else:  # we are on Mac; sqlite3 module doesn't work in Mac IronPython
        # Execute the honybee CLI to obtain the results via CPython
        cmds = [folders.python_exe_path, '-m', 'honeybee_energy', 'result',
                'data-by-outputs', sql]
        for outp in energy_output:
            cmds.append('["{}"]'.format(outp))
        custom_env = os.environ.copy()
        custom_env['PYTHONHOME'] = ''
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, env=custom_env)
        stdout = process.communicate()
        data_coll_dicts = json.loads(stdout[0])
        ac_power = serialize_data(data_coll_dicts[0])
        dc_power = serialize_data(data_coll_dicts[1])

    # give a warning if no generators were found in the results
    if len(ac_power) == 0:
        msg = 'No electricity generation objects were found in the connected _hb_objs.\n' \
            'Try applying PV properties to some of the connected Shade geometries.'
        print(msg)
        give_warning(ghenv.Component, msg)

    # group the generator results by identifier
    dc_dict = OrderedDict()
    for g_data in dc_power:
        gen_id = g_data.header.metadata['System'].split('..')[0]
        g_data.header.metadata['System'] = gen_id
        try:
            dc_dict[gen_id] += g_data
        except KeyError:
            dc_dict[gen_id] = g_data

    # sum the results and ouput them
    total_ac = [acp.total for acp in ac_power]
    generators = list(dc_dict.keys())
    total_dc = [dcp.total for dcp in dc_dict.values()]
    dc_power = [dcp for dcp in dc_dict.values()]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Set the properites of a Model's electric load center, which governs how on-site
electricity generation is converted and distributed.
-

    Args:
        _model: A Honeybee Model for which the electric load center properties will be set.
            This Model should include on-site power generation objects, like
            Shades with PV properties assigned, in order for the inputs here
            to have an effect on the simulation.
        _inverter_eff_: A number between 0 and 1 for the load centers's inverter nominal
            rated DC-to-AC conversion efficiency. An inverter converts DC power,
            such as that output by photovoltaic panels, to AC power, such as
            that distributed by the electrical grid and is available from
            standard electrical outlets. Inverter efficiency is defined as the
            inverter's rated AC power output divided by its rated DC power
            output. (Default: 0.96).
        _dc_to_ac_size_: A positive number (typically greater than 1) for the ratio of the
            inverter's DC rated size to its AC rated size. Typically, inverters
            are not sized to convert the full DC output under standard test
            conditions (STC) as such conditions rarely occur in reality and
            therefore unnecessarily add to the size/cost of the inverter. For a
            system with a high DC to AC size ratio, during times when the 
            DC power output exceeds the inverter's rated DC input size, the inverter
            limits the array's power output by increasing the DC operating voltage,
            which moves the arrays operating point down its current-voltage (I-V)
            curve. The default value of 1.1 is reasonable for most systems. A
            typical range is 1.1 to 1.25, although some large-scale systems have
            ratios of as high as 1.5. The optimal value depends on the system's
            location, array orientation, and module cost. (Default: 1.1).

    Returns:
        model: The input Honeybee Model with the electric load center properties set.
"""

ghenv.Component.Name = 'HB Electric Load Center'
ghenv.Component.NickName = 'ElectricCenter'
ghenv.Component.Message = '1.7.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '4 :: HVAC'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the input Model to avoid editing it
    model = _model.duplicate()

    # set default properties
    _inverter_eff_ = 0.96 if _inverter_eff_ is None else _inverter_eff_
    _dc_to_ac_size_ = 1.1 if _dc_to_ac_size_ is None else _dc_to_ac_size_
    model.properties.energy.electric_load_center.inverter_efficiency = _inverter_eff_
    model.properties.energy.electric_load_center.inverter_dc_to_ac_size_ratio = _inverter_eff_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a Setpoint object that can be used to create a ProgramType or be assigned
directly to a Room.
-

    Args:
        _name_: Text to set the name for the Setpoint and to be incorporated
            into a unique Setpoint identifier. If None, a unique name will
            be generated.
         _heating_sch: A temperature schedule for the heating setpoint.
            The type limit of this schedule should be temperature and the
            values should be the temperature setpoint in degrees Celcius.
        _cooling_sch: A temperature schedule for the cooling setpoint.
            The type limit of this schedule should be temperature and the
            values should be the temperature setpoint in degrees Celcius.
        humid_setpt_: A numerical value between 0 and 100 for the relative humidity
            humidifying setpoint [%]. This value will be constant throughout the
            year. If None, no humidification will occur.
        dehumid_setpt_: A numerical value between 0 and 100 for the relative humidity
            dehumidifying setpoint [%]. This value will be constant throughout the
            year. If None, no dehumidification will occur beyond that which is needed
            to create air at the cooling supply temperature.
        cutout_difference_: An optional positive number for the temperature difference between the
            cutout temperature and the setpoint temperature. Specifying a non-zero
            number here is useful for modeling the throttling range associated
            with a given setup of setpoint controls and HVAC equipment. Throttling
            ranges describe the range where a zone is slightly over-cooled or
            over-heated beyond the thermostat setpoint. They are used to avoid
            situations where HVAC systems turn on only to turn off a few minutes later,
            thereby wearing out the parts of mechanical systems faster. They can
            have a minor impact on energy consumption and can often have significant
            impacts on occupant thermal comfort, though using the default value
            of zero will often yield results that are close enough when trying
            to estimate the annual heating/cooling energy use. Specifying a value
            of zero effectively assumes that the system will turn on whenever
            conditions are outside the setpoint range and will cut out as soon
            as the setpoint is reached. (Default: 0).

    Returns:
        setpoint: A Setpoint object that can be used to create a ProgramType or
            be assigned directly to a Room.
"""

ghenv.Component.Name = 'HB Setpoint'
ghenv.Component.NickName = 'Setpoint'
ghenv.Component.Message = '1.7.1'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee_energy.load.setpoint import Setpoint
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # make a default Setpoint name if none is provided
    name = clean_and_id_ep_string('Setpoint') if _name_ is None else \
        clean_ep_string(_name_)

    # get the schedules
    if isinstance(_heating_sch, str):
        _heating_sch = schedule_by_identifier(_heating_sch)
    if isinstance(_cooling_sch, str):
        _cooling_sch = schedule_by_identifier(_cooling_sch)

    # create the Setpoint object
    setpoint = Setpoint(name, _heating_sch, _cooling_sch)
    if _name_ is not None:
        setpoint.display_name = _name_

    # assign the humidification and dehumidification setpoints if specified
    if humid_setpt_ is not None:
        setpoint.humidifying_setpoint = humid_setpt_
    if dehumid_setpt_ is not None:
        setpoint.dehumidifying_setpoint = dehumid_setpt_

    # assign the cutout_difference_ if specified
    if cutout_difference_ is not None:
        setpoint.setpoint_cutout_difference = cutout_difference_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Add ventilation fans to Rooms.
_
This fan is not connected to any heating or cooling system and is meant to
represent the intentional circulation of unconditioned outdoor air for the
purposes of keeping a space cooler, drier or free of indoor pollutants (as in
the case of kitchen or bathroom exhaust fans).
-

    Args:
        _rooms: Honeybee Rooms to which ventilation fans should be assigned.
        _name_: Text to set the name for the ventilation fan and to be incorporated into a
            unique ventilation fan identifier. If None, a unique name will be
            generated.
        _flow_rate: A positive number for the flow rate of the fan in m3/s.
        _vent_type_: Text to indicate the type of type of ventilation. Choose from the
            options below. For either Exhaust or Intake, values for fan pressure
            and efficiency define the fan electric consumption. For Exhaust
            ventilation, the conditions of the air entering the space are assumed
            to be equivalent to outside air conditions. For Intake and Balanced
            ventilation, an appropriate amount of fan heat is added to the
            entering air stream. For Balanced ventilation, both an intake fan
            and an exhaust fan are assumed to co-exist, both having the same
            flow rate and power consumption (using the entered values for fan
            pressure rise and fan total efficiency). Thus, the fan electric
            consumption for Balanced ventilation is twice that for the Exhaust or 
            Intake ventilation types which employ only a single fan. (Default: Balanced).
                * Exhaust
                * Intake
                * Balanced
        _pressure_rise_: A number for the the pressure rise across the fan in Pascals (N/m2).
            This is often a function of the fan speed and the conditions in
            which the fan is operating since having the fan blow air through filters
            or narrow ducts will increase the pressure rise that is needed to
            deliver the input flow rate. The pressure rise plays an important role in
            determining the amount of energy consumed by the fan. Smaller fans like
            a 0.05 m3/s desk fan tend to have lower pressure rises around 60 Pa.
            Larger fans, such as a 6 m3/s fan used for ventilating a large room tend
            to have higher pressure rises around 400 Pa. The highest pressure rises
            are typically for large fans blowing air through ducts and filters, which
            can have pressure rises as high as 1000 Pa. If this input is None,
            the pressure rise will be estimated from the flow_rate, with higher
            flow rates corresponding to larger pressure rises (up to 400 Pa). These
            estimated pressure rises are generally assumed to have minimal obstructions
            between the fan and the room and they should be increased if the fan is
            blowing air through ducts or filters.
        _efficiency_: A number between 0 and 1 for the overall efficiency of the fan.
            Specifically, this is the ratio of the power delivered to the fluid
            to the electrical input power. It is the product of the fan motor
            efficiency and the fan impeller efficiency. Fans that have a higher blade
            diameter and operate at lower speeds with smaller pressure rises for
            their size tend to have higher efficiencies. Because motor efficiencies
            are typically between 0.8 and 0.9, the best overall fan efficiencies
            tend to be around 0.7 with most typical fan efficiencies between 0.5 and
            0.7. The lowest efficiencies often happen for small fans in situations
            with high pressure rises for their size, which can result in efficiencies
            as low as 0.15. If None, this input will be estimated from the fan
            flow rate and pressure rise with large fans operating at low pressure
            rises for their size having up to 0.7 efficiency and small fans
            operating at high pressure rises for their size having as low as
            0.15 efficiency.
        vent_cntrl_: A Ventilation Control object from the "HB Ventilation Control"
            component that dictates the conditions under which the
            fan is turned on. If None, the fan on all of the time.

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with ventilation fans assigned to them.
"""

ghenv.Component.Name = 'HB Fan Ventilation'
ghenv.Component.NickName = 'FanVent'
ghenv.Component.Message = '1.7.0'
ghenv.Component.Category = 'HB-Energy'
ghenv.Component.SubCategory = '3 :: Loads'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import the honeybee extension
    from honeybee.typing import clean_and_id_ep_string, clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.ventcool.fan import VentilationFan
    from honeybee_energy.lib.schedules import schedule_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, longest_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # duplicate the initial objects
    rooms = [room.duplicate() for room in _rooms]

    # set default values and check the inputs
    _vent_type_ = ['Balanced'] if len(_vent_type_) == 0 else _vent_type_
    _pressure_rise_ = [None] if len(_pressure_rise_) == 0 else _pressure_rise_
    _efficiency_ = [None] if len(_efficiency_) == 0 else _efficiency_
    vent_cntrl_ = [None] if len(vent_cntrl_) == 0 else vent_cntrl_

    # loop through the rooms and assign ventilation fans
    for i, room in enumerate(rooms):
        flow_rate = longest_list(_flow_rate, i)
        if flow_rate != 0:
            name = clean_and_id_ep_string('VentilationFan') if len(_name_) == 0 else \
                clean_ep_string(longest_list(_name_, i))
            fan = VentilationFan(
                name, flow_rate,  longest_list(_vent_type_, i),
                longest_list(_pressure_rise_, i),
                longest_list(_efficiency_, i), longest_list(vent_cntrl_, i)
            )
            room.properties.energy.add_fan(fan)
